//////////////////////////////////////////////////////////////////////////
// $Id: Connection.cpp,v 1.19 2005/09/01 20:49:17 klas Exp $
//
// Usage: Data class. inherited from Connection, everything pertaining to
//        data connection goes here.
//
//  A part of babyftpd, licensed under the GNU GPL, see the file
//    LICENSE for complete information.
//////////////////////////////////////////////////////////////////////////

#define CONNECTION_CPP
#define NETINCLS
#define BABY_CORE
#include "Connection.h"

Connection::Connection()
{
  pthread_rwlock_init(&this->lock, NULL);
  this->_settings.reset();
  this->_packet_size = READ_SIZE;
  this->_transfered = 0;
}

Connection::~Connection()
{
  if(this->_socket_id)
    close(this->_socket_id);

  pthread_rwlock_destroy(&this->lock);

#ifdef USE_TLS
  // TLS cleanup here.
  if(this->settings(TLS_INITED))
  {
    gnutls_db_remove_session(this->session);
    gnutls_deinit(this->session);
  }
#endif
  if(this->thread_id())
  {
    sem_wait(&cleanup_list_lock);
    cleanup_list.push_back(this->thread_id());
    sem_post(&cleanup_list_lock);
  }
}

ssize_t Connection::send(const char *message, size_t len)
{
  assert(len <= this->_packet_size && len > 0);
  ssize_t ret = -1;

  BABY_WRLOCK(this->lock);
#ifdef USE_TLS
  if(this->settings(TLS_ENABLED))
  {
    ret = gnutls_record_send(this->session, message, len);
    if(ret == GNUTLS_E_AGAIN || ret == GNUTLS_E_INTERRUPTED)
      ret = gnutls_record_send(this->session, message, len);
  }
  else
#endif // USE_TLS
    ret = write(this->_socket_id, message, len);

  BABY_UNLOCK(this->lock);
  if(ret < 0)
    logging->log_this(3, TYPE_INFO, "write to socket " +
        util.itos(this->socket_id()) + " failed: %m");
  else
    this->transfered(this->transfered() + ret);

  return ret;
}

ssize_t Connection::receive(char *message, size_t len)
{
  assert(len <= this->_packet_size && len > 0);
  ssize_t ret = -1;

#ifdef USE_TLS
  if(this->settings(TLS_ENABLED))
    ret = gnutls_record_recv(this->session, message, len);
  else
#endif
    ret = read(this->socket_id(), message, len);

  this->transfered(this->transfered() + ret);

  return ret;
}

Connection &Connection::operator<<(long long int num)
{
  return (*this << util.itos(num));
}

Connection &Connection::operator<<(string message)
{
  if(this->send(message.c_str(), message.size()) < 0)
  this->settings(FAILURE, B_SET);
  return *this;
}

#ifdef USE_TLS
bool Connection::engage_tls(void)
{
  bool ret = false;
  gnutls_init(&session, GNUTLS_SERVER);

  gnutls_set_default_priority(this->session);

  gnutls_credentials_set(this->session, GNUTLS_CRD_CERTIFICATE, x509_cred);

  gnutls_certificate_server_set_request(this->session, GNUTLS_CERT_REQUEST);

  gnutls_dh_set_prime_bits(this->session, 1024);

  // these are for the connection cache
  gnutls_db_set_retrieve_function(this->session, tls_db_fetch);
  gnutls_db_set_remove_function(this->session, tls_db_delete);
  gnutls_db_set_store_function(this->session, tls_db_store);
  gnutls_db_set_ptr(this->session, NULL);

  // associating the session with the socket.
  gnutls_transport_set_ptr(this->session,
      reinterpret_cast<gnutls_transport_ptr>(this->socket_id()));
  this->settings(TLS_INITED, B_SET);

  if(gnutls_handshake(this->session) >= 0)
  {
    this->settings(TLS_ENABLED, B_SET);
    this->packet_size(gnutls_record_get_max_size(this->session));
    ret = true;
  }

  return ret;
}

bool Connection::disengage_tls(void)
{
  gnutls_bye(this->session, GNUTLS_SHUT_RDWR);
  this->settings(TLS_ENABLED, B_RESET);
  return true;
}

bool Connection::verify_cert(void)
{
  bool ret = false;
  if (gnutls_certificate_type_get(this->session) == GNUTLS_CRT_X509)
  {
    int verified = gnutls_certificate_verify_peers(this->session);
    if(util.is_set(verified, GNUTLS_CERT_REVOKED))
      logging->log_this(2, TYPE_AUTH, this->user_name() +
          + " tried to login with a revoked certificate.");
    else if(!util.is_set(verified, GNUTLS_CERT_INVALID) &&
        !util.is_set(verified, GNUTLS_CERT_SIGNER_NOT_FOUND))
    {
      unsigned int cert_list_size = 0;
      // I really hope gnutls has their locking code in shape. get_peers just
      // gives us some memory.. *shudder*                     
      const gnutls_datum *cert_list = gnutls_certificate_get_peers(
          this->session, &cert_list_size);
      if(cert_list_size > 0)
      {
        gnutls_x509_crt peer_cert;
        gnutls_x509_crt_init(&peer_cert);
        gnutls_x509_crt_import(peer_cert, &cert_list[0], GNUTLS_X509_FMT_DER);
        char dn_buf[200] = {0}, ca_buf[200] = {0};
        size_t buf_size = 199, ca_size = 199;
        gnutls_x509_crt_get_dn(peer_cert, dn_buf, &buf_size);
        gnutls_x509_crt_get_issuer_dn(peer_cert, ca_buf, &ca_size);
        if(buf_size > 0 && ca_size > 0)
        {
          // here we call an external program to ask if this user may login
          // with this cert.
          //  format is "<user> <host/ip> <ca-dn> <cert-dn>"
          string call = config->general.verify_client + " " +
            this->user_name() + " " + this->host_name() +
            " \"" + ca_buf + "\" \"" + dn_buf + "\"";
          logging->log_this(5, TYPE_DEBUG, "running:" + call);
          int approved = system(call.c_str());
          if(WEXITSTATUS(approved) == 0)
            ret = true;
        }
        gnutls_x509_crt_deinit(peer_cert);
      }
    }
  }
  return ret;
}
#endif // USE_TLS
