///////////////////////////////////////////////////////////////////
//  $Id: Log.cpp,v 1.32 2005/08/23 19:34:02 klas Exp $
//
//  Usage: Logging functions etc.
//
//  A part of babyftpd, licensed under the GNU GPL, see the file
//    LICENSE for complete information.
/////////////////////////////////////////////////////////////////////

#define LOG_CPP
#define BABY_IO
#include "Log.h"

class Log *logging;

Log::Log(void)
{
  sem_init(&this->to_log_lock, 0, 1);
  pthread_cond_init(&data_ready, NULL);
  pthread_mutex_init(&ready_mutex, NULL);

  openlog("BabyFTPd", LOG_PID, LOG_FTP);
}

Log::~Log(void)
{
  closelog();

  sem_destroy(&this->to_log_lock);
  pthread_cond_destroy(&this->data_ready);
  pthread_mutex_destroy(&this->ready_mutex);
}

void Log::log_this(int log_level, enum log_types log_type, string log_message)
{
  if(log_level > config->logger.log_level && log_type != TYPE_XFER)
    return; // no point in pushing data that wont be logged..

  struct log_data temp_data = {log_level, log_type, log_message};

  if(temp_data.type != TYPE_XFER && config->logger.thread_id)
    temp_data.message += " (" + util.itos(pthread_self()) + ")";

  sem_wait(&this->to_log_lock);
  this->to_log.push_back(temp_data);
  sem_post(&this->to_log_lock);

  // bump data_ready to tell main thread there is data
  pthread_mutex_lock(&this->ready_mutex);
  pthread_cond_signal(&this->data_ready);
  pthread_mutex_unlock(&this->ready_mutex);
}


void Log::log_main(void)
{
  setuid(config->general.user);
  setgid(config->general.group);

  size_t time_string_len;

  {
    string temp;
    time_string_len = util.time_string(temp, config->logger.time_format, 500);
    time_string_len++;
  }

  while(true)
  {
    pthread_mutex_lock(&this->ready_mutex);
    while(this->to_log.empty())
      pthread_cond_wait(&this->data_ready, &this->ready_mutex);
    pthread_mutex_unlock(&this->ready_mutex);

    sem_wait(&this->to_log_lock);
    struct log_data temp_data = this->to_log.front();
    this->to_log.erase(this->to_log.begin());
    sem_post(&this->to_log_lock);

    switch(temp_data.level)
    {
      case 0:
        temp_data.level = LOG_CRIT;
        break;

      case 1:
        temp_data.level = LOG_ERR;
        break;

      case 2:
        temp_data.level = LOG_WARNING;
        break;

      case 3:
        temp_data.level = LOG_NOTICE;
        break;

      case 4:
        temp_data.level = LOG_INFO;
        break;

      case 5:
      default: // so we get everything else as well ;)
        temp_data.level = LOG_DEBUG;
        break;
    }

    if((temp_data.type == TYPE_AUTH && !config->logger.authlog.empty()) ||
        (temp_data.type == TYPE_INFO && !config->logger.infolog.empty()) ||
        (temp_data.type == TYPE_DEBUG && !config->logger.debuglog.empty()))
    {
      string timestr;
      util.time_string(timestr, config->logger.time_format, time_string_len);
      temp_data.message.insert(0, timestr);
    }

    switch(temp_data.type)
    {
      case TYPE_XFER:
        this->xferlog(temp_data);
        break;

      case TYPE_AUTH:
        this->authlog(temp_data);
        break;

      case TYPE_INFO:
        this->infolog(temp_data);
        break;

      case TYPE_DEBUG:
        this->debuglog(temp_data);
        break;
    }
  }
}

void Log::xferlog(struct log_data data_to_log)
{
  ofstream xstream;
  xstream.open(config->logger.xferlog_file.c_str(),
      ofstream::out | ofstream::app);
  if(!xstream)
    log_this(2, TYPE_INFO, "Failed to open xferlog file \""
        + config->logger.xferlog_file  + "\" for writing, check permissions.");
  else
  {
    xstream << data_to_log.message << endl;
    xstream.close();
  }
}

void Log::authlog(struct log_data data_to_log)
{
  if(config->logger.authlog.empty())
    syslog(LOG_AUTHPRIV | data_to_log.level, data_to_log.message.c_str());
  else
  {
    ofstream astream;
    astream.open(config->logger.authlog.c_str(),
        ofstream::out | ofstream::app);
    if(!astream)
      log_this(2, TYPE_INFO, "Failed to open authlog file \""
          + config->logger.authlog  + "\" for writing, check permissions.");
    else
    {
      astream << data_to_log.message << endl;
      astream.close();
    }
  }
}

void Log::infolog(struct log_data data_to_log)
{
  if(config->logger.infolog.empty())
    syslog(LOG_FTP | data_to_log.level, data_to_log.message.c_str());
  else
  {
    ofstream istream;
    istream.open(config->logger.infolog.c_str(),
        ofstream::out | ofstream::app);
    if(!istream)
      log_this(2, TYPE_INFO, "Failed to open infolog file \""
          + config->logger.infolog  + "\" for writing, check permissions.");
    else
    {
      istream << data_to_log.message << endl;
      istream.close();
    }
  }
}

void Log::debuglog(struct log_data data_to_log)
{
  if(config->logger.infolog.empty())
    syslog(LOG_FTP | data_to_log.level, data_to_log.message.c_str());
  else
  {
    ofstream dstream;
    dstream.open(config->logger.debuglog.c_str(),
        ofstream::out | ofstream::app);
    if(!dstream)
      log_this(2, TYPE_INFO, "Failed to open debuglog file \""
          + config->logger.debuglog  + "\" for writing, check permissions.");
    else
    {
      dstream << data_to_log.message << endl;
      dstream.close();
    }
  }
}
